/* traj.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/* trajectory tracing module */



#include <math.h>
#include "globals.h"
#include "grid.h"




static float xlonf[1000], xlatf, xhgtf;




void init_traj()
{
   int i;
   float rdpdg, xmpdg, xlat;

   /* calculate xlonf, xlatf, xhgtf */
   rdpdg = 3.1415926/180.0;
   xmpdg = 111137.0;
   for (i=0;i<Nr;i++) {
      xlat = NorthLat[0] - i * LatInc;
      xlonf[i] = 1.0 / (xmpdg*LonInc*cos(rdpdg*xlat));
   }
   xlatf = 1.0 / (xmpdg*LatInc);
   xhgtf = 0.001 / HgtInc;

}




#define GET_U0(R,C,L)  (  ((val=ucomp0[((L)*Nc+(C))*Nr+(R)])==127) ? 1.0e35 \
                              : ( ((float)val-ugb0)*ugc0 + ugr0)*xlonf[R]  )

#define GET_V0(R,C,L)  (  ((val=vcomp0[((L)*Nc+(C))*Nr+(R)])==127) ? 1.0e35 \
                              : ( ((float)val-vgb0)*vgc0 + vgr0)*xlatf  )

#define GET_W0(R,C,L)  (  ((val=wcomp0[((L)*Nc+(C))*Nr+(R)])==127) ? 1.0e35 \
                              : ( ((float)val-wgb0)*wgc0 + wgr0)*xhgtf  )

#define GET_U1(R,C,L)  (  ((val=ucomp1[((L)*Nc+(C))*Nr+(R)])==127) ? 1.0e35 \
                              : ( ((float)val-ugb1)*ugc1 + ugr1)*xlonf[R]  )

#define GET_V1(R,C,L)  (  ((val=vcomp1[((L)*Nc+(C))*Nr+(R)])==127) ? 1.0e35 \
                              : ( ((float)val-vgb1)*vgc1 + vgr1)*xlatf  )

#define GET_W1(R,C,L)  (  ((val=wcomp1[((L)*Nc+(C))*Nr+(R)])==127) ? 1.0e35 \
                              : ( ((float)val-wgb1)*wgc1 + wgr1)*xhgtf  )


#define ABS(A)   ((A)<0.0 ? -(A) : (A))
#define MAX 2000

#define STOP    { inbox = 0;  break; }


/*** trace ************************************************************
   Trace a wind trajectory.
   Input:  curr, curc, curl - init. pos. in [0..Nr-1],[0..Nc-1],[0..Nl-1]
           ctime            - init. time in [0..NumTimes-1]
           step             - step size (~100.0)
	   len              - trajectory length multiplier (1.0)
           max              - max number of vertices to return
   Output:  vr, vc, vl      - array of vertices in grid coords
            vt              - array of vertex times
	    ifrlen          - used to calc tstart and tlen.
   Return:  number of vertices put into arrays.
**********************************************************************/
int trace( curr, curc, curl, ctime, step, len,   max, vr, vc, vl, vt, ifrlen )
float curr, curc, curl;
int ctime;
float step, len;
int max;
float vr[], vc[], vl[];
int vt[];
int *ifrlen;
{
   float xentrj;
   int inext[MAX], itint[MAX];
   float zr[MAX], zc[MAX], zl[MAX];
   double delsec, etime;
   float zlim, xfact, zd, clat;
   int jtint[MAX], iday, itim;
   int lint, ktime, kint, ktint, mint;
   int itobs, itm, nints, irev;
   int lsint, lday, mtrajv, lensec, jjday, jd;
   int ilist, le, lt, lrat, ltime, ls, jjtime;
   int i,j,k;
   float qr, qc, ql;
   float stotal, sint, stold;
   float zra, zca, zla, zrz, zcz, zlz;
   float at, bt;
   float bra, ara, bca, aca, bla, ala, aa, ba, ab, bb;
   float aaaa,abaa,baaa,bbaa,aaba,abba,baba,bbba,
         aaab,abab,baab,bbab,aabb,abbb,babb,bbbb;
   float ullll, ullhl, ulhll, ulhhl, uhlll, uhlhl, uhhll, uhhhl,
         ulllh, ullhh, ulhlh, ulhhh, uhllh, uhlhh, uhhlh, uhhhh;
   float vllll, vllhl, vlhll, vlhhl, vhlll, vhlhl, vhhll, vhhhl,
         vlllh, vllhh, vlhlh, vlhhh, vhllh, vhlhh, vhhlh, vhhhh;
   float wllll, wllhl, wlhll, wlhhl, whlll, whlhl, whhll, whhhl,
         wlllh, wllhh, wlhlh, wlhhh, whllh, whlhh, whhlh, whhhh;
   float uuu, vvv, www;
   int_1 *ucomp0, *vcomp0, *wcomp0, *ucomp1, *vcomp1, *wcomp1, val;
   float uga0, ugb0, ugc0, ugr0;
   float uga1, ugb1, ugc1, ugr1;
   float vga0, vgb0, vgc0, vgr0;
   float vga1, vgb1, vgc1, vgr1;
   float wga0, wgb0, wgc0, wgr0;
   float wga1, wgb1, wgc1, wgr1;
   int inbox;

   if (max>MAX)
      max = MAX;

   ctime++;
   xentrj = len;

   /* ZLIM IS MINIMUM DISTANCE (IN BOXES) BETWEEN RECORDED POINTS */
   zlim = (float) (Nc-1) / step;     /* ZLIM=(MC-1.0)/STEP */

   /* CALCULATE LINT SO THAT A 25M/S WIND MOVES APPROX ZLIM IN A STEP */
   /* NOT CORRECTED FOR LON SHORTENING WITH LAT */
   /* 44.45=111137.0/(100.0*25.0)=(M/DEG)/(100*(M/SEC)) */
   xfact = 111137.0/(step*25.0);
   lint = (WestLon[0]-EastLon[0])*xfact;
   if (lint<=0)
      lint = 1;

   /* ADJUST LINT TO BE NO MORE THAN A TENTH OF THE TOTAL TIME */
   /*  86400=24*60*60 */
   /*     KTIME IS TOTAL TIME SPANNED ACROSS THE GRIDS IN SECONDS */
   ktime = 86400*(DayStamp[NumTimes-1]-DayStamp[0])
          +TimeStamp[NumTimes-1]-TimeStamp[0];
   kint = ktime/10;
   if (kint < lint)
     lint = kint;

   /* REAL GRID DIMENSIONS */
   qr = (float) (Nr-1);
   qc = (float) (Nc-1);
   ql = (float) (Nl-1);

   /*  USE CURSOR POSITION AND CURSOR TIME TO INITIALIZE THE */
   /*  TRAJECTORY LIST. */
   zr[0] = zra = zrz = curr;
   zc[0] = zca = zcz = curc;
   zl[0] = zla = zlz = curl;
   itint[0] = 0;
   /*printf("1. itint[0] = 0\n");*/
   inext[0] = -1;
   ilist = 0;
   le = 0;

   /* START TRACING TRAJECTORY FORWARD */
   /* IF A TRAJECTORY LEAVES THE BOX, GOES BELOW THE TOPOGRAPHY, */
   /* OR ENCOUNTERS MISSING DATA, IT WILL BE TERMINATED. */

   /* LOOP OVER TIME FRAMES FROM CURSOR-TIME TO LAST TIME FRAME */

   inbox = 1;
   for (itobs=ctime; itobs<NumTimes && inbox; itobs++) {
      /* DO 190 ITOBS=CTIME+1,NTIMES */

      /* CALCULATE NUMBER OF TIME STEPS BETWEEN DATA SETS */
      itm = Elapsed[itobs] - Elapsed[itobs-1];
      nints = (itm+lint-1)/lint;
      if (nints <= 0)
	 nints = 1;
      stotal = itm;
      sint = lint;

      /* get the u,v,w grids for itobs-1 and itobs */
      ucomp0 = get_compressed_grid( itobs-1, Uparm, &uga0, &ugb0 );
      ugc0 = 1.0 / uga0;
      ugr0 = 0.0001 * 250.0 * ugc0;
      vcomp0 = get_compressed_grid( itobs-1, Vparm, &vga0, &vgb0 );
      vgc0 = 1.0 / vga0;
      vgr0 = 0.0001 * 250.0 * vgc0;
      wcomp0 = get_compressed_grid( itobs-1, Wparm, &wga0, &wgb0 );
      wgc0 = 1.0 / wga0;
      wgr0 = 0.0001 * 250.0 * wgc0;
      ucomp1 = get_compressed_grid( itobs, Uparm, &uga1, &ugb1 );
      ugc1 = 1.0 / uga1;
      ugr1 = 0.0001 * 250.0 * ugc1;
      vcomp1 = get_compressed_grid( itobs, Vparm, &vga1, &vgb1 );
      vgc1 = 1.0 / vga1;
      vgr1 = 0.0001 * 250.0 * vgc1;
      wcomp1 = get_compressed_grid( itobs, Wparm, &wga1, &wgb1 );
      wgc1 = 1.0 / wga1;
      wgr1 = 0.0001 * 250.0 * wgc1;

      /* ITERATE BETWEEN TIME FRAMES TO PRODUCE 'SMOOTH' TRAJECTORY */

      for (lt=1;lt<=nints && inbox;lt++) {
	 /* DO 175 LT=1,NINTS */
	 if (stotal < sint)
	    sint = stotal;
	 stold = stotal;
	 stotal = stotal-sint;
	 lsint = sint;

	 /* CALCULATE WEIGHTS FOR INTERPOLATING BETWEEN SETS */
	 at = (stotal+stold)/(2.0*itm);
	 bt = 1.0-at;

	 /* CHECK FOR TRAJECTORY OUT OF BOX, IF SO, START TO TRACE BACKWARD */
	 if (zra < 0.0 || zra > qr) STOP
	 if (zca < 0.0 || zca > qc) STOP
	 if (zla < 0.0 || zla > ql) STOP

	 /* CALCULATE WEIGHTS FOR INTERPOLATING IN BOX */
	 i = (int) zra;
	 j = (int) zca;
	 k = (int) zla;

	 bra = zra-i;
	 ara = 1.0-bra;
	 bca = zca-j;
	 aca = 1.0-bca;
	 bla = zla-k;
	 ala = 1.0-bla;

	 aa = ala*at;
	 ba = bla*at;
	 ab = ala*bt;
	 bb = bla*bt;

	 /* GET 16 WEIGHTS FOR 4-D BOX */
	 /* NOTE 4 A/B'S ARE IN ORDER OF J,I,K,TIME */
	 aaaa = aca*ara*aa;
	 abaa = aca*bra*aa;
	 baaa = bca*ara*aa;
	 bbaa = bca*bra*aa;
	 aaba = aca*ara*ba;
	 abba = aca*bra*ba;
	 baba = bca*ara*ba;
	 bbba = bca*bra*ba;

	 aaab = aca*ara*ab;
	 abab = aca*bra*ab;
	 baab = bca*ara*ab;
	 bbab = bca*bra*ab;
	 aabb = aca*ara*bb;
	 abbb = aca*bra*bb;
	 babb = bca*ara*bb;
	 bbbb = bca*bra*bb;

	 /* GET U GRID VALUES AT 16 CORNERS OF 4-D BOX */
	 /* NOTE 4 L/H'S ARE IN ORDER OF J,I,K,TIME */
	 ullll = GET_U0(i,j,k);
	 ullhl = GET_U0(i,j,k+1);
	 ulhll = GET_U0(i+1,j,k);
	 ulhhl = GET_U0(i+1,j,k+1);
	 uhlll = GET_U0(i,j+1,k);
	 uhlhl = GET_U0(i,j+1,k+1);
	 uhhll = GET_U0(i+1,j+1,k);
	 uhhhl = GET_U0(i+1,j+1,k+1);
	 ulllh = GET_U1(i,j,k);
	 ullhh = GET_U1(i,j,k+1);
	 ulhlh = GET_U1(i+1,j,k);
	 ulhhh = GET_U1(i+1,j,k+1);
	 uhllh = GET_U1(i,j+1,k);
	 uhlhh = GET_U1(i,j+1,k+1);
	 uhhlh = GET_U1(i+1,j+1,k);
	 uhhhh = GET_U1(i+1,j+1,k+1);
	 /* CHECK FOR MISSING DATA */
	 if (ullll>1.e30 || ullhl>1.e30 || ulhll>1.e30 || ulhhl>1.e30) STOP;
	 if (uhlll>1.e30 || uhlhl>1.e30 || uhhll>1.e30 || uhhhl>1.e30) STOP;
	 if (ulllh>1.e30 || ullhh>1.e30 || ulhlh>1.e30 || ulhhh>1.e30) STOP;
	 if (uhllh>1.e30 || uhlhh>1.e30 || uhhlh>1.e30 || uhhhh>1.e30) STOP;

	 /* GET V GRID VALUES AT 16 CORNERS OF 4-D BOX */
	 /* NOTE 4 L/H'S ARE IN ORDER OF J,I,K,TIME */
	 vllll = GET_V0(i,j,k);
	 vllhl = GET_V0(i,j,k+1);
	 vlhll = GET_V0(i+1,j,k);
	 vlhhl = GET_V0(i+1,j,k+1);
	 vhlll = GET_V0(i,j+1,k);
	 vhlhl = GET_V0(i,j+1,k+1);
	 vhhll = GET_V0(i+1,j+1,k);
	 vhhhl = GET_V0(i+1,j+1,k+1);
	 vlllh = GET_V1(i,j,k);
	 vllhh = GET_V1(i,j,k+1);
	 vlhlh = GET_V1(i+1,j,k);
	 vlhhh = GET_V1(i+1,j,k+1);
	 vhllh = GET_V1(i,j+1,k);
	 vhlhh = GET_V1(i,j+1,k+1);
	 vhhlh = GET_V1(i+1,j+1,k);
	 vhhhh = GET_V1(i+1,j+1,k+1);
	 /* check for missing data */
	 if (vllll>1.e30 || vllhl>1.e30 || vlhll>1.e30 || vlhhl>1.e30) STOP;
	 if (vhlll>1.e30 || vhlhl>1.e30 || vhhll>1.e30 || vhhhl>1.e30) STOP;
	 if (vlllh>1.e30 || vllhh>1.e30 || vlhlh>1.e30 || vlhhh>1.e30) STOP;
	 if (vhllh>1.e30 || vhlhh>1.e30 || vhhlh>1.e30 || vhhhh>1.e30) STOP;

	 /* get w grid values at 16 corners of 4-d box */
	 /* note 4 l/h's are in order of j,i,k,time */
	 wllll = GET_W0(i,j,k);
	 wllhl = GET_W0(i,j,k+1);
	 wlhll = GET_W0(i+1,j,k);
	 wlhhl = GET_W0(i+1,j,k+1);
	 whlll = GET_W0(i,j+1,k);
	 whlhl = GET_W0(i,j+1,k+1);
	 whhll = GET_W0(i+1,j+1,k);
	 whhhl = GET_W0(i+1,j+1,k+1);
	 wlllh = GET_W1(i,j,k);
	 wllhh = GET_W1(i,j,k+1);
	 wlhlh = GET_W1(i+1,j,k);
	 wlhhh = GET_W1(i+1,j,k+1);
	 whllh = GET_W1(i,j+1,k);
	 whlhh = GET_W1(i,j+1,k+1);
	 whhlh = GET_W1(i+1,j+1,k);
	 whhhh = GET_W1(i+1,j+1,k+1);
	 /*  check for missing data */
	 if (wllll>1.e30 || wllhl>1.e30 || wlhll>1.e30 || wlhhl>1.e30) STOP;
	 if (whlll>1.e30 || whlhl>1.e30 || whhll>1.e30 || whhhl>1.e30) STOP;
	 if (wlllh>1.e30 || wllhh>1.e30 || wlhlh>1.e30 || wlhhh>1.e30) STOP;
	 if (whllh>1.e30 || whlhh>1.e30 || whhlh>1.e30 || whhhh>1.e30) STOP;

	 /* interpolate to get wind vector (uuu,vvv,www) */
	 uuu = aaaa*ullll+aaab*ulllh+aaba*ullhl+aabb*ullhh+
	       abaa*ulhll+abab*ulhlh+abba*ulhhl+abbb*ulhhh+
	       baaa*uhlll+baab*uhllh+baba*uhlhl+babb*uhlhh+
	       bbaa*uhhll+bbab*uhhlh+bbba*uhhhl+bbbb*uhhhh;
	 vvv = aaaa*vllll+aaab*vlllh+aaba*vllhl+aabb*vllhh+
	       abaa*vlhll+abab*vlhlh+abba*vlhhl+abbb*vlhhh+
	       baaa*vhlll+baab*vhllh+baba*vhlhl+babb*vhlhh+
	       bbaa*vhhll+bbab*vhhlh+bbba*vhhhl+bbbb*vhhhh;
	 www = aaaa*wllll+aaab*wlllh+aaba*wllhl+aabb*wllhh+
	       abaa*wlhll+abab*wlhlh+abba*wlhhl+abbb*wlhhh+
               baaa*whlll+baab*whllh+baba*whlhl+babb*whlhh+
               bbaa*whhll+bbab*whhlh+bbba*whhhl+bbbb*whhhh;

	 /* MULTIPLY BY SINT TO CONVERT BOXES/SEC TO BOXES/TIME-STEP */
	 /* POSITIVE WINDS ARE FROM SOUTH, WEST AND BELOW */
	 zra = zra-vvv*sint;
	 zca = zca+uuu*sint;
	 zla = zla+www*sint;

	 /* CHECK FOR TRAJECTORY BELOW TOPOGRAPHY */

	 /* ACCUMULATE TIME STEP BETWEEN RECORDED POINTS */
	 itint[le] = itint[le]+lsint;
	 /*printf("2. itint[%d] = %d\n", le, itint[le] );*/
	 ktint = itint[le]+lint;

	 /* DON'T RECORD TRAJ POINT IF TOO CLOSE TO LAST RECORDED POINT */

	 zd =sqrt((zra-zrz)*(zra-zrz)+(zca-zcz)*(zca-zcz)+(zla-zlz)*(zla-zlz));
	 if (zd < zlim && ktint < 30000)
	    continue; /*don't record*/

	 /* RECORD POINT */
	 zrz = zra;
	 zcz = zca;
	 zlz = zla;
	 le = ilist;
	 ilist = ilist + 1;
	 zr[ilist] = zra;
	 zc[ilist] = zca;
	 zl[ilist] = zla;
	 itint[ilist] = 0;
	 /*printf("3. itint[%d] = %d\n", ilist, itint[ilist] );*/
	 inext[le] = ilist;
	 if (ilist == max-1)
	    STOP;

	 /* END OF LOOP FOR ITERATING BETWEEN TIME FRAMES */
      }  /*175*/
      release_compressed_grid( itobs-1, Uparm );
      release_compressed_grid( itobs-1, Vparm );
      release_compressed_grid( itobs-1, Wparm );
      release_compressed_grid( itobs, Uparm );
      release_compressed_grid( itobs, Vparm );
      release_compressed_grid( itobs, Wparm );

   } /*190*/


   /* NOW GO THROUGH TIME STEPS BACKWARD: */
   zra = zrz = curr;
   zca = zcz = curc;
   zla = zlz = curl;
   mint = 0;
   irev = ilist;

   /* LOOP BACKWARD OVER TIME FRAMES FROM CTIME-1 TO 2 */
   inbox = 1;

   for (itobs=ctime-1-1;itobs>=0 && inbox;itobs--) {
      /*  DO 290 ITOBS=CTIME-1,1,-1 */

      /* CALCULATE NUMBER OF TIME STEPS BETWEEN DATA SETS */
      itm = Elapsed[itobs+1]-Elapsed[itobs];
      nints = (itm+lint-1) / lint;
      if(nints <= 0) nints = 1;
      stotal = -itm;
      sint = -lint;

      /* get the u,v,w grids for itobs+1 and itobs */
      ucomp0 = get_compressed_grid( itobs+1, Uparm, &uga0, &ugb0 );
      ugc0 = 1.0 / uga0;
      ugr0 = 0.0001 * 250.0 * ugc0;
      vcomp0 = get_compressed_grid( itobs+1, Vparm, &vga0, &vgb0 );
      vgc0 = 1.0 / vga0;
      vgr0 = 0.0001 * 250.0 * vgc0;
      wcomp0 = get_compressed_grid( itobs+1, Wparm, &wga0, &wgb0 );
      wgc0 = 1.0 / wga0;
      wgr0 = 0.0001 * 250.0 * wgc0;
      ucomp1 = get_compressed_grid( itobs, Uparm, &uga1, &ugb1 );
      ugc1 = 1.0 / uga1;
      ugr1 = 0.0001 * 250.0 * ugc1;
      vcomp1 = get_compressed_grid( itobs, Vparm, &vga1, &vgb1 );
      vgc1 = 1.0 / vga1;
      vgr1 = 0.0001 * 250.0 * vgc1;
      wcomp1 = get_compressed_grid( itobs, Wparm, &wga1, &wgb1 );
      wgc1 = 1.0 / wga1;
      wgr1 = 0.0001 * 250.0 * wgc1;

      /* ITERATE OVER THE INTERVAL BETWEEN TIME FRAMES */
      for (lt=1;lt<=nints && inbox;lt++) {
	 /* DO 275 LT=1,NINTS */

	 if (stotal > sint)
	   sint = stotal;
	 stold = stotal;
	 stotal = stotal-sint;
	 lsint = -sint;

	 /* CALCULATE WEIGHTS FOR INTERPOLATING BETWEEN SETS */
	 at = -(stotal+stold)/(2.0*itm);
	 bt = 1.0-at;

	 /* CHECK FOR TRAJECTORY OUT OF BOX */
	 if (zra < 0.0 || zra > qr) STOP;
	 if (zca < 0.0 || zca > qc) STOP;
	 if (zla < 0.0 || zla > ql) STOP;

	 /* CALCULATE WEIGHTS FOR INTERPOLATING IN BOX */
	 i = (int) zra;
	 j = (int) zca;
	 k = (int) zla;

	 bra = zra-i;
	 ara = 1.0-bra;
	 bca = zca-j;
	 aca = 1.0-bca;
	 bla = zla-k;
	 ala = 1.0-bla;

	 aa = ala*at;
	 ba = bla*at;
	 ab = ala*bt;
	 bb = bla*bt;

	 /* GET 16 WEIGHTS FOR 4-D BOX */
	 /* NOTE 4 A/B'S ARE IN ORDER OF J,I,K,TIME */
	 aaaa = aca*ara*aa;
	 abaa = aca*bra*aa;
	 baaa = bca*ara*aa;
	 bbaa = bca*bra*aa;
	 aaba = aca*ara*ba;
	 abba = aca*bra*ba;
	 baba = bca*ara*ba;
	 bbba = bca*bra*ba;

	 aaab = aca*ara*ab;
	 abab = aca*bra*ab;
	 baab = bca*ara*ab;
	 bbab = bca*bra*ab;
	 aabb = aca*ara*bb;
	 abbb = aca*bra*bb;
	 babb = bca*ara*bb;
	 bbbb = bca*bra*bb;

	 /* GET U GRID VALUES AT 16 CORNERS OF 4-D BOX */
	 /* NOTE 4 L/H'S ARE IN ORDER OF J,I,K,TIME */
	 ullll = GET_U0(i,j,k);
	 ullhl = GET_U0(i,j,k+1);
	 ulhll = GET_U0(i+1,j,k);
	 ulhhl = GET_U0(i+1,j,k+1);
	 uhlll = GET_U0(i,j+1,k);
	 uhlhl = GET_U0(i,j+1,k+1);
	 uhhll = GET_U0(i+1,j+1,k);
	 uhhhl = GET_U0(i+1,j+1,k+1);
	 ulllh = GET_U1(i,j,k);
	 ullhh = GET_U1(i,j,k+1);
	 ulhlh = GET_U1(i+1,j,k);
	 ulhhh = GET_U1(i+1,j,k+1);
	 uhllh = GET_U1(i,j+1,k);
	 uhlhh = GET_U1(i,j+1,k+1);
	 uhhlh = GET_U1(i+1,j+1,k);
	 uhhhh = GET_U1(i+1,j+1,k+1);
	 /* CHECK FOR MISSING DATA */
	 if (ullll>1.e30 || ullhl>1.e30 || ulhll>1.e30 || ulhhl>1.e30) STOP;
	 if (uhlll>1.e30 || uhlhl>1.e30 || uhhll>1.e30 || uhhhl>1.e30) STOP;
	 if (ulllh>1.e30 || ullhh>1.e30 || ulhlh>1.e30 || ulhhh>1.e30) STOP;
	 if (uhllh>1.e30 || uhlhh>1.e30 || uhhlh>1.e30 || uhhhh>1.e30) STOP;

	 /* GET V GRID VALUES AT 16 CORNERS OF 4-D BOX */
	 /* NOTE 4 L/H'S ARE IN ORDER OF J,I,K,TIME */
	 vllll = GET_V0(i,j,k);
	 vllhl = GET_V0(i,j,k+1);
	 vlhll = GET_V0(i+1,j,k);
	 vlhhl = GET_V0(i+1,j,k+1);
	 vhlll = GET_V0(i,j+1,k);
	 vhlhl = GET_V0(i,j+1,k+1);
	 vhhll = GET_V0(i+1,j+1,k);
	 vhhhl = GET_V0(i+1,j+1,k+1);
	 vlllh = GET_V1(i,j,k);
	 vllhh = GET_V1(i,j,k+1);
	 vlhlh = GET_V1(i+1,j,k);
	 vlhhh = GET_V1(i+1,j,k+1);
	 vhllh = GET_V1(i,j+1,k);
	 vhlhh = GET_V1(i,j+1,k+1);
	 vhhlh = GET_V1(i+1,j+1,k);
	 vhhhh = GET_V1(i+1,j+1,k+1);
	 /* CHECK FOR MISSING DATA */
	 if (vllll>1.e30 || vllhl>1.e30 || vlhll>1.e30 || vlhhl>1.e30) STOP;
	 if (vhlll>1.e30 || vhlhl>1.e30 || vhhll>1.e30 || vhhhl>1.e30) STOP;
	 if (vlllh>1.e30 || vllhh>1.e30 || vlhlh>1.e30 || vlhhh>1.e30) STOP;
	 if (vhllh>1.e30 || vhlhh>1.e30 || vhhlh>1.e30 || vhhhh>1.e30) STOP;

	 /* GET W GRID VALUES AT 16 CORNERS OF 4-D BOX */
	 /* NOTE 4 L/H'S ARE IN ORDER OF J,I,K,TIME */
	 wllll = GET_W0(i,j,k);
	 wllhl = GET_W0(i,j,k+1);
	 wlhll = GET_W0(i+1,j,k);
	 wlhhl = GET_W0(i+1,j,k+1);
	 whlll = GET_W0(i,j+1,k);
	 whlhl = GET_W0(i,j+1,k+1);
	 whhll = GET_W0(i+1,j+1,k);
	 whhhl = GET_W0(i+1,j+1,k+1);
	 wlllh = GET_W1(i,j,k);
	 wllhh = GET_W1(i,j,k+1);
	 wlhlh = GET_W1(i+1,j,k);
	 wlhhh = GET_W1(i+1,j,k+1);
	 whllh = GET_W1(i,j+1,k);
	 whlhh = GET_W1(i,j+1,k+1);
	 whhlh = GET_W1(i+1,j+1,k);
	 whhhh = GET_W1(i+1,j+1,k+1);
	 /* CHECK FOR MISSING DATA */
	 if (wllll>1.e30 || wllhl>1.e30 || wlhll>1.e30 || wlhhl>1.e30) STOP;
	 if (whlll>1.e30 || whlhl>1.e30 || whhll>1.e30 || whhhl>1.e30) STOP;
	 if (wlllh>1.e30 || wllhh>1.e30 || wlhlh>1.e30 || wlhhh>1.e30) STOP;
	 if (whllh>1.e30 || whlhh>1.e30 || whhlh>1.e30 || whhhh>1.e30) STOP;

	 /* INTERPOLATE TO GET WIND VECTOR (UUU,VVV,WWW) */
	 uuu = aaaa*ullll+aaab*ulllh+aaba*ullhl+aabb*ullhh+
	       abaa*ulhll+abab*ulhlh+abba*ulhhl+abbb*ulhhh+
	       baaa*uhlll+baab*uhllh+baba*uhlhl+babb*uhlhh+
	       bbaa*uhhll+bbab*uhhlh+bbba*uhhhl+bbbb*uhhhh;
	 vvv = aaaa*vllll+aaab*vlllh+aaba*vllhl+aabb*vllhh+
	       abaa*vlhll+abab*vlhlh+abba*vlhhl+abbb*vlhhh+
	       baaa*vhlll+baab*vhllh+baba*vhlhl+babb*vhlhh+
	       bbaa*vhhll+bbab*vhhlh+bbba*vhhhl+bbbb*vhhhh;
	 www = aaaa*wllll+aaab*wlllh+aaba*wllhl+aabb*wllhh+
	       abaa*wlhll+abab*wlhlh+abba*wlhhl+abbb*wlhhh+
	       baaa*whlll+baab*whllh+baba*whlhl+babb*whlhh+
	       bbaa*whhll+bbab*whhlh+bbba*whhhl+bbbb*whhhh;

	 /* MULTIPLY BY SINT TO CONVERT BOXES/SEC TO BOXES/TIME-STEP */
	 /* POSITIVE WINDS FROM SOUTH, WEST AND BELOW */
	 zra = zra-vvv*sint;
	 zca = zca+uuu*sint;
	 zla = zla+www*sint;

	 /* CHECK FOR TRAJECTORY BELOW TOPOGRAPHY */
	 
	 /* ACCUMULATE TIME STEP BETWEEN RECORDED POINTS */
	 mint = mint+lsint;
	 ktint = mint+lint;
	 /* DON'T RECORD TRAJ POINT IF TOO CLOSE TO LAST RECORDED POINT */
	 zd =sqrt((zra-zrz)*(zra-zrz)+(zca-zcz)*(zca-zcz)+(zla-zlz)*(zla-zlz));
	 if (zd < zlim && ktint < 30000)
	    continue;

	 /* TIME IN SECONDS SINCE MIDNIGHT */
	 lrat = lt * lint;
	 if (lrat > itm)
	    lrat = itm;
	 ltime = Elapsed[itobs+1] - lrat;
	 if (ltime < 0)
	    ltime = 0;

	 /* RECORD POINT - NOTE IT IS RECORDED BEFORE OLD FIRST POINT */
	 zrz = zra;
	 zcz = zca;
	 zlz = zla;
	 itint[ilist] = mint;
	 /*printf("4. itint[%d] = %d\n", ilist, itint[ilist] );*/
	 mint = 0;
	 ls = ilist;
	 ilist = ilist + 1;
	 zr[ilist] = zra;
	 zc[ilist] = zca;
	 zl[ilist] = zla;
	 inext[ilist] = ls;
	 if (ilist == max-1)  STOP;

	 /* END OF LOOP BETWEEN GRID TIME FRAMES */
      } /*275*/
      release_compressed_grid( itobs+1, Uparm );
      release_compressed_grid( itobs+1, Vparm );
      release_compressed_grid( itobs+1, Wparm );
      release_compressed_grid( itobs, Uparm );
      release_compressed_grid( itobs, Vparm );
      release_compressed_grid( itobs, Wparm );

      /* END OF LOOP OVER GRID TIME FRAMES */

   } /*290*/


   /**** DONE TRACING THE TRAJECTORY ****/

   itint[ilist] = 0;
   /*printf("5. itint[%d] = %d\n", ilist, itint[ilist] );*/

   /* UPDATE THE START DAY AND TIME (LDAY AND LTIME) FOR THIS TRAJ */
   jjtime = ltime + TimeStamp[0];
   k = jjtime/86400;
   ltime = jjtime-k*86400;
   lday = DayStamp[0] + k;

   /* PATCH UP THE LINKED LIST and copy vertices to vr,vc,vl */
   inext[irev] = -1;
   inext[irev+1] = 0;

   if (irev == ilist)
     i = 0;
   else
     i = ilist;

   j = 0;
   while (i!=-1 && j<max) {
      vr[j] = zr[i];
      vc[j] = zc[i];
      vl[j] = zl[i];
      jtint[j] = itint[i];
      /*printf("jtint[%d] = %d\n", j, jtint[j] );*/
      if (itint[i]<0 || itint[i]>=100000)
	 printf("BAD ITINT[%d]=%d\n", i, itint[i] );
      i = inext[i];
      j = j + 1;
   }
/*
   for (i=0;i<30;i++) {
      printf("inext[%d] = %d\n", i, inext[i] );
   }
*/
   mtrajv = j;

   /* CONVERT TRAJECTORIES TO RIBBONS HERE */
/*   if (ribbon) { */
      /* ktraj = 1 */
      /* IF (IRIB > 0) CALL RIBBON(KTRAJ,KTRAJ) */
/*   } */

   /* CALCULATE TIME STEP IN SECONDS */
   /*     (A WIND OF 132 M/S WILL DRAW A TRAJ ACROSS THE BOX) */
   clat = cos(0.5*(NorthLat[0]+SouthLat[0]));
   lensec = xentrj*(111137.0/132.0)*ABS(clat*(WestLon[0]-EastLon[0]));
   delsec = ( (DayStamp[NumTimes-1]-DayStamp[0])*86400
	     + TimeStamp[NumTimes-1]-TimeStamp[0] )/(NumTimes-1);
   *ifrlen = lensec/delsec;
   if (*ifrlen < 1)
      *ifrlen=1;

   /* CALCULATE vt() VALUES */

   iday = lday;
   itim = ltime;
   for (i=0;i<mtrajv;i++) {
      jjday = iday;
      jjtime = itim;
      jd = jjtime / 86400;
      jjtime = jjtime - jd*86400;
      jjday = jjday + jd;
      /* -calculate time (in secs) from beginning to current traj time */
      etime = (jjday-DayStamp[0])*86400 + jjtime-TimeStamp[0];
      /* -calculate animation time step (head & tail) */
      vt[i] = (int) ( (etime/delsec) + 0.5 );
      /*printf("vt[%d]=%d %f %f %f\n", i, vt[i], vr[i], vc[i], vl[i] );*/
      itim = itim + jtint[i];
   }

   return mtrajv;
}



#define EPS 0.0000000001
#define WSCALE 150.0
#define RADIUS 10.0



int to_ribbon( num, xt, yt, zt, itint, xn,yn,zn )
int num;
float xt[], yt[], zt[];
int itint[];
float xn[], yn[], zn[];
{
   int i, it, ip, il, ih;
   int iuu, imm, idd, iu, id;
   float ax, ay, az, bx, by, bz, cx, cy, cz, dx, dy, dz, ex, ey, ez;
   float cc, dd, ee;
   float scale;
   float xr[MAX], yr[MAX], zr[MAX];
   int itinr[MAX];

   scale = 2.0/WSCALE;

   for (i=0;i<num;i++) {
      xr[i] = xt[i];
      yr[i] = yt[i];
      zr[i] = zt[i];
      itinr[i] = itint[i];
   }

   /* DO 300 ITRAJ=ITRLOW,ITRHI */
   /* IL = LSTART(ITRAJ) */
   /* IH = LEND(ITRAJ) */
   /* CALCULATE UNFILTERED RIBBON ORIENTATIONS */
   il = 0;
   ih = num;
   for (i=il;i<ih;i++) {
      /* DO 10 I=IL,IH */
      it = 2*i - 0;
      ip = it+1;

      /* SHORT TRAJECTORY CASE */
      if (ih-il <= 2) {
	 xt[it] = xr[i];
	 yt[it] = yr[i];
	 zt[it] = zr[i];
	 xn[it] = 1.0;
	 yn[it] = 0.0;
	 zn[it] = 0.0;
	 itint[it] = 0;
	 xt[ip] = xr[i];
	 yt[ip] = yr[i];
	 zt[ip] = zr[i];
	 xn[ip] = 1.0;
	 yn[ip] = 0.0;
	 zn[ip] = 0.0;
	 itint[ip] = itinr[i];

	 /* go to 10 */
      }
      else {
	 /* LONG TRAJECTORY CASE */
	 /* RAW CROSS RIBBON VECTORS */
	 iuu = i+1;
	 imm = i;
	 idd = i-1;
	 if (i == il) {
	    iuu = i+2;
	    imm = i+1;
	    idd = i;
	 }
	 if (i == ih-1) {
	    iuu = i;
	    imm = i-1;
	    idd = i-2;
	 }
	 ax = xr[iuu]-xr[imm];
	 ay = yr[iuu]-yr[imm];
	 az = zr[iuu]-zr[imm];
	 bx = xr[imm]-xr[idd];
	 by = yr[imm]-yr[idd];
	 bz = zr[imm]-zr[idd];
	 cx = ay*bz-az*by;
	 cy = az*bx-ax*bz;
	 cz = ax*by-ay*bx;
	 cc = sqrt(cx*cx+cy*cy+cz*cz);
	 if (cc < EPS)
	   cc = EPS;
	 cc = 1.0/cc;
	 xn[ip] = cx*cc;
	 yn[ip] = cy*cc;
	 zn[ip] = cz*cc;
	 /* ALONG RIBBON VECTOR */
	 iu = i+1;
	 id = i-1;
	 if (i == il) id = i;
	 if (i == ih-1) iu = i;
	 dx = xr[iu]-xr[id];
	 dy = yr[iu]-yr[id];
	 dz = zr[iu]-zr[id];
	 dd = sqrt(dx*dx+dy*dy+dz*dz);
	 if (dd < EPS)
	   dd = EPS;
	 dd = 1.0/dd;
	 xn[it] = dx*dd;
	 yn[it] = dy*dd;
	 zn[it] = dz*dd;
      }
   }
   /* 10    CONTINUE */

   if (ih-il < 2)   /* <= 2 ??? */
      return 0;

   /* PROPOGATE CROSS RIBBON VECTOR */
   cx = 0.0;
   cy = 0.0;
   cz = 0.0;

   /* DO 20 I=IL,IH */
   for (i=il;i<ih;i++) {
      it = 2*i-0;
      ip = it+1;
      /* ADD (OR SUBTRACT) RAW CROSS RIBBON FROM RUNNING SUM */
      ax=xn[ip];
      ay=yn[ip];
      az=zn[ip];
      if (ax*cx+ay*cy+az*cz > 0.0) {
	 cx=cx+ax;
	 cy=cy+ay;
	 cz=cz+az;
      }
      else {
	 cx=cx-ax;
	 cy=cy-ay;
	 cz=cz-az;
      }

      /* GET COMPONENT OF (CROSS RIBBON) NORMAL TO ALONG RIBBON */
      dx=xn[it];
      dy=yn[it];
      dz=zn[it];
      dd=cx*dx+cy*dy+cz*dz;
      cx=cx-dd*dx;
      cy=cy-dd*dy;
      cz=cz-dd*dz;
      /* SCALE CROSS RIBBON TO RADIUS LENGTH */
      cc = sqrt(cx*cx+cy*cy+cz*cz);
      if (cc < EPS)
	cc = EPS;
      cc = RADIUS/cc;
      cx=cx*cc;
      cy=cy*cc;
      cz=cz*cc;
      xt[it]=cx;
      yt[it]=cy;
      zt[it]=cz;
   }
   /* 20    CONTINUE; */

   /* NOW DO RIBBONS */
   /* DO 40 I=IL,IH */
   for (i=il;i<ih;i++) {
      it = 2*i - 0;
      ip = it+1;
      cx=xt[it];
      cy=yt[it];
      cz=zt[it];
      dx=xn[it];
      dy=yn[it];
      dz=zn[it];

      cc = scale/RADIUS;
      cx=cx*cc;
      cy=cy*cc;
      cz=cz*cc;
      /* CALCULATE NORMAL TO RIBBON */
      ex = cy*dz - cz*dy;
      ey = cz*dx - cx*dz;
      ez = cx*dy - cy*dx;
      ee = sqrt(ex*ex+ey*ey+ez*ez);
      if (ee < EPS) ee = EPS;
      ee = 1.0/ee;
      ex=ex*ee;
      ey=ey*ee;
      ez=ez*ee;
      /* MAKE RIBBON POINTS */
      xt[it]=xr[i] - cx;
      yt[it]=yr[i] - cy;
      zt[it]=zr[i] - cz;
      xn[it]=ex;
      yn[it]=ey;
      zn[it]=ez;
      itint[it]=0;
      xt[ip]=xr[i] + cx;
      yt[ip]=yr[i] + cy;
      zt[ip]=zr[i] + cz;
      xn[ip]=ex;
      yn[ip]=ey;
      zn[ip]=ez;
      itint[ip]=itinr[i];
   }
   /* 40    CONTINUE */

   return num*2;
}
